@@ -1,4 +1,4 @@
-from dataclasses import Field, dataclass, field, is_dataclass, replace
+from dataclasses import Field, dataclass, is_dataclass
 from enum import Enum
 from functools import wraps
 from itertools import chain
@@ -6,7 +6,6 @@ from typing import (
     AbstractSet,
     Any,
     Callable,
-    Collection,
     Dict,
     Iterable,
     List,
@@ -17,6 +16,7 @@ from typing import (
     Tuple,
     Type,
     TypeVar,
+    Union,
     cast,
     overload,
 )
@@ -24,7 +24,12 @@ from typing import (
 from apischema import settings
 from apischema.aliases import Aliaser
 from apischema.cache import cache
-from apischema.conversions.utils import Conversions, ConversionsWrapper, identity
+from apischema.conversions.conversions import (
+    Conversions,
+    HashableConversions,
+    to_hashable_conversions,
+)
+from apischema.conversions.utils import Converter, identity
 from apischema.conversions.visitor import Deserialization, DeserializationVisitor
 from apischema.dataclass_utils import (
     get_alias,
@@ -35,6 +40,7 @@ from apischema.dataclass_utils import (
     has_default,
     is_required,
 )
+from apischema.dataclasses import replace
 from apischema.dependent_required import DependentRequired
 from apischema.deserialization.coercion import Coercion, get_coercer
 from apischema.deserialization.merged import get_init_merged_alias
@@ -45,6 +51,7 @@ from apischema.json_schema.constraints import (
 )
 from apischema.json_schema.patterns import infer_pattern
 from apischema.json_schema.schema import Schema, get_schema
+from apischema.metadata.implem import ValidatorsMetadata
 from apischema.metadata.keys import (
     DEFAULT_FALLBACK_METADATA,
     MERGED_METADATA,
@@ -65,7 +72,7 @@ from apischema.types import (
     OrderedDict,
 )
 from apischema.typing import get_origin
-from apischema.utils import Operation
+from apischema.utils import OperationKind, opt_or
 from apischema.validation.errors import (
     ErrorKey,
     ValidationError,
@@ -75,7 +82,6 @@ from apischema.validation.errors import (
 from apischema.validation.mock import ValidatorMock
 from apischema.validation.validator import (
     Validator,
-    ValidatorsMetadata,
     get_validators,
     validate,
 )
@@ -93,18 +99,37 @@ T = TypeVar("T")
 # TODO maybe ctx parameters to deserializers
 @dataclass
 class DeserializationContext:
-    additional_properties: bool = field(
-        default_factory=lambda: settings.additional_properties
-    )
-    coercion: Coercion = field(default_factory=lambda: settings.coercion)
-    default_fallback: bool = field(default_factory=lambda: settings.default_fallback)
+    additional_properties: bool
+    coercion: Coercion
+    default_fallback: bool
 
     def __post_init__(self):
         self.coercer = get_coercer(self.coercion)
 
+    def merge(
+        self,
+        additional_properties: bool = None,
+        coercion: Coercion = None,
+        default_fallback: bool = None,
+    ) -> "DeserializationContext":
+        if any(
+            arg is not None
+            for arg in (additional_properties, coercion, default_fallback)
+        ):
+            return replace(
+                self,
+                additional_properties=opt_or(
+                    additional_properties, self.additional_properties
+                ),
+                coercion=opt_or(coercion, self.coercion),
+                default_fallback=opt_or(default_fallback, self.default_fallback),
+            )
+        else:
+            return self
 
-def get_constraints(cls: AnyType) -> Optional[Constraints]:
-    schema = get_schema(cls)
+
+def get_constraints(tp: AnyType) -> Optional[Constraints]:
+    schema = get_schema(tp)
     return schema.constraints if schema is not None else None
 
 
@@ -123,9 +148,9 @@ class DeserializationMethodFactory:
         return self.factory(self.constraints, self.validators)  # type: ignore
 
     @staticmethod
-    def from_type(cls: AnyType) -> Callable[[Factory], "DeserializationMethodFactory"]:
+    def from_type(tp: AnyType) -> Callable[[Factory], "DeserializationMethodFactory"]:
         return lambda factory: DeserializationMethodFactory(
-            factory, get_constraints(cls), get_validators(cls)
+            factory, get_constraints(tp), get_validators(tp)
         )
 
     def merge(
@@ -232,47 +257,6 @@ def field_deserializer(
     return deserializer
 
 
-RequiredFieldChecker = Callable[
-    [Mapping[str, Any], Dict[ErrorKey, ValidationError]], None
-]
-
-
-def required_field_checker(
-    field: Field, requiring: Optional[Collection[Field]], aliaser: Aliaser
-) -> RequiredFieldChecker:
-    alias = aliaser(get_alias(field))
-    required = is_required(field)
-    if requiring is not None:
-        requiring_alias = {aliaser(get_alias(req)) for req in requiring}
-
-        def checker(
-            data: Mapping[str, Any], field_errors: Dict[ErrorKey, ValidationError]
-        ):
-            required_by = requiring_alias & data.keys()
-            if required_by:
-                field_errors[alias] = ValidationError(
-                    [f"missing property (required by {sorted(required_by)})"]
-                )
-            elif required:
-                field_errors[alias] = MISSING_PROPERTY
-
-    elif required:
-
-        def checker(
-            data: Mapping[str, Any], field_errors: Dict[ErrorKey, ValidationError]
-        ):
-            field_errors[alias] = MISSING_PROPERTY
-
-    else:
-
-        def checker(
-            data: Mapping[str, Any], field_errors: Dict[ErrorKey, ValidationError]
-        ):
-            pass
-
-    return checker
-
-
 def with_validators(
     validators: Sequence[Validator],
 ) -> Callable[[DeserializationMethod], DeserializationMethod]:
@@ -297,22 +281,22 @@ class DeserializationMethodVisitor(
         self._rec_sentinel: Dict[Any, RecDeserializerMethodFactory] = {}
         self.aliaser = aliaser
 
-    def _visit(self, cls: AnyType) -> DeserializationMethodFactory:
-        key = self._resolve_type_vars(cls), id(self._conversions)
+    def _visit(self, tp: AnyType) -> DeserializationMethodFactory:
+        key = self._generic or tp, self._conversions
         if key in self._rec_sentinel:
             return cast(DeserializationMethodFactory, self._rec_sentinel[key])
         else:
             self._rec_sentinel[key] = RecDeserializerMethodFactory()
-            factory = super()._visit(cls)
+            factory = super()._visit(tp)
             return self._rec_sentinel.pop(key).set_ref(factory)
 
     def method(self, cls) -> DeserializationMethod:
         return self.visit(cls).method
 
     def annotated(
-        self, cls: AnyType, annotations: Sequence[Any]
+        self, tp: AnyType, annotations: Sequence[Any]
     ) -> DeserializationMethodFactory:
-        factory = self.visit(cls)
+        factory = self.visit(tp)
         for annotation in reversed(annotations):
             if isinstance(annotation, Schema):
                 factory = factory.merge(constraints=annotation.constraints)
@@ -370,7 +354,9 @@ class DeserializationMethodVisitor(
         init_vars: Sequence[Field],
     ) -> DeserializationMethodFactory:
         assert is_dataclass(cls)
-        normal_fields: List[Tuple[str, FieldDeserializer, RequiredFieldChecker]] = []
+        normal_fields: List[
+            Tuple[str, FieldDeserializer, Union[bool, AbstractSet[str]]]
+        ] = []
         merged_fields: List[Tuple[AbstractSet[str], FieldDeserializer]] = []
         pattern_fields: List[Tuple[Pattern, FieldDeserializer]] = []
         additional_field: Optional[FieldDeserializer] = None
@@ -381,26 +367,22 @@ class DeserializationMethodVisitor(
         }
         defaults: Dict[str, Callable[[], Any]] = {}
         required_by = get_requirements(
-            cls, DependentRequired.required_by, Operation.DESERIALIZATION
+            cls, DependentRequired.required_by, OperationKind.DESERIALIZATION
         )
         for field in get_fields(  # noqa: F402
-            fields, init_vars, Operation.DESERIALIZATION
+            fields, init_vars, OperationKind.DESERIALIZATION
         ):
             metadata = check_metadata(field)
             if SKIP_METADATA in metadata or not field.init:
                 continue
-            field_type = types[field.name]
             if has_default(field):
                 defaults[field.name] = lambda: get_default(field)
-            conversion_type, conversions, converter = get_field_conversion(
-                field, field_type, Operation.DESERIALIZATION
+            field_type = types[field.name]
+            field_type, conversion = get_field_conversion(
+                field, field_type, OperationKind.DESERIALIZATION
             )
-            if converter is not None:
-                field_factory = self.visit_with_conversions(
-                    field_type, {conversion_type: (converter, conversions)}
-                )
-            elif conversions is not None:
-                field_factory = self.visit_with_conversions(field_type, conversions)
+            if conversion is not None:
+                field_factory = self.visit_conversion(field_type, [conversion])
             else:
                 field_factory = self.visit(field_type)
             if SCHEMA_METADATA in metadata:
@@ -426,11 +408,11 @@ class DeserializationMethodVisitor(
                 else:
                     pattern_fields.append((pattern, deserializer))
             else:
-                checker = required_field_checker(
-                    field, required_by.get(field), self.aliaser
-                )
+                required = is_required(field) or {
+                    self.aliaser(get_alias(req)) for req in required_by.get(field) or ()
+                }
                 normal_fields.append(
-                    (self.aliaser(get_alias(field)), deserializer, checker)
+                    (self.aliaser(get_alias(field)), deserializer, required)
                 )
 
         @DeserializationMethodFactory.from_type(cls)
@@ -444,14 +426,23 @@ class DeserializationMethodVisitor(
                 errors = [] if constraints is None else constraints.errors(data)
                 field_errors: Dict[ErrorKey, ValidationError] = OrderedDict()
 
-                for alias, deserialize_field, required_checker in normal_fields:
+                for alias, deserialize_field, required in normal_fields:
                     if alias in data:
                         aliases.append(alias)
                         deserialize_field(
                             ctx, data[alias], values, errors, field_errors
                         )
+                    elif not required:
+                        pass
+                    elif required is True:
+                        field_errors[alias] = MISSING_PROPERTY
                     else:
-                        required_checker(data, field_errors)
+                        assert isinstance(required, AbstractSet)
+                        requiring = required & data.keys()
+                        if requiring:
+                            msg = f"missing property (required by {sorted(requiring)})"
+                            field_errors[alias] = ValidationError([msg])
+
                 for merged_alias, deserialize_field in merged_fields:
                     merged = {
                         alias: data[alias] for alias in merged_alias if alias in data
@@ -663,9 +654,9 @@ class DeserializationMethodVisitor(
         return factory
 
     def new_type(
-        self, cls: AnyType, super_type: AnyType
+        self, tp: AnyType, super_type: AnyType
     ) -> DeserializationMethodFactory:
-        return self.visit(super_type).merge(get_constraints(cls), get_validators(cls))
+        return self.visit(super_type).merge(get_constraints(tp), get_validators(tp))
 
     def primitive(self, cls: Type) -> DeserializationMethodFactory:
         @DeserializationMethodFactory
@@ -799,12 +790,18 @@ class DeserializationMethodVisitor(
         return factory
 
     def visit_conversion(
-        self, cls: AnyType, conversion: Deserialization
+        self, cls: Type, conversion: Deserialization
     ) -> DeserializationMethodFactory:
         assert conversion
         factories = [
-            (self.visit_with_conversions(source, conversions), converter)
-            for source, (converter, conversions) in conversion.items()
+            (
+                self.visit_with_conversions(
+                    self._replace_generic_args(conv.source), conv.conversions
+                ),
+                cast(Converter, conv.converter),
+                (conv.additional_properties, conv.coercion, conv.default_fallback),
+            )
+            for conv in conversion
         ]
 
         @DeserializationMethodFactory.from_type(cls)
@@ -812,18 +809,20 @@ class DeserializationMethodVisitor(
             constraints: Optional[Constraints], validators: Sequence[Validator]
         ) -> DeserializationMethod:
             alt_deserializers = [
-                (fact.merge(constraints, validators).method, converter)
-                for fact, converter in factories
+                (fact.merge(constraints, validators).method, converter, conv_ctx)
+                for fact, converter, conv_ctx in factories
             ]
             if len(alt_deserializers) == 1:
-                ((deserialize_alt, converter),) = alt_deserializers
+                deserialize_alt, converter, conv_ctx = alt_deserializers[0]
                 if converter is identity:
                     method = deserialize_alt
                 else:
 
                     def method(ctx: DeserializationContext, data: Any) -> Any:
                         try:
-                            return converter(deserialize_alt(ctx, data))
+                            return converter(
+                                deserialize_alt(ctx.merge(*conv_ctx), data)
+                            )
                         except (ValidationError, AssertionError):
                             raise
                         except Exception as err:
@@ -833,9 +832,9 @@ class DeserializationMethodVisitor(
 
                 def method(ctx: DeserializationContext, data: Any) -> Any:
                     error: Optional[ValidationError] = None
-                    for deserialize_alt, converter in alt_deserializers:
+                    for deserialize_alt, converter, conv_ctx in alt_deserializers:
                         try:
-                            value = deserialize_alt(ctx, data)
+                            value = deserialize_alt(ctx.merge(*conv_ctx), data)
                             break
                         except ValidationError as err:
                             error = merge_errors(error, err)
@@ -856,20 +855,17 @@ class DeserializationMethodVisitor(
 
 @cache
 def get_method(
-    cls: AnyType,
-    wrapper: Optional[ConversionsWrapper],
-    aliaser: Aliaser,
+    tp: AnyType, conversions: Optional[HashableConversions], aliaser: Aliaser
 ) -> DeserializationMethod:
-    conversions = wrapper.conversions if wrapper is not None else None
     factory = DeserializationMethodVisitor(aliaser).visit_with_conversions(
-        cls, conversions
+        tp, conversions
     )
     return factory.method
 
 
 @overload
 def deserialize(
-    cls: Type[T],
+    tp: Type[T],
     data: Any,
     *,
     conversions: Conversions = None,
@@ -883,7 +879,7 @@ def deserialize(
 
 @overload
 def deserialize(
-    cls: AnyType,
+    tp: AnyType,
     data: Any,
     *,
     conversions: Conversions = None,
@@ -896,7 +892,7 @@ def deserialize(
 
 
 def deserialize(
-    cls: AnyType,
+    tp: AnyType,
     data: Any,
     *,
     conversions: Conversions = None,
@@ -914,5 +910,4 @@ def deserialize(
     if aliaser is None:
         aliaser = settings.aliaser()
     ctx = DeserializationContext(additional_properties, coercion, default_fallback)
-    wrapper = ConversionsWrapper(conversions) if conversions is not None else None
-    return get_method(cls, wrapper, aliaser)(ctx, data)
+    return get_method(tp, to_hashable_conversions(conversions), aliaser)(ctx, data)
