@@ -1,5 +1,3 @@
-import warnings
-from contextlib import suppress
 from dataclasses import Field, InitVar, dataclass, field as field_
 from enum import Enum
 from inspect import Parameter, iscoroutinefunction
@@ -26,7 +24,7 @@ from typing import (
 
 import graphql
 
-from apischema import serialize
+from apischema import UndefinedType, serialize
 from apischema.aliases import Aliaser
 from apischema.conversions import identity
 from apischema.conversions.conversions import Conversions
@@ -122,19 +120,17 @@ class ObjectField:
     name: str
     type: AnyType
     alias: Optional[str] = None
+    converter: Converter = identity
     conversions: Optional[Conversions] = None
     default: Any = graphql.Undefined
     parameters: Optional[Tuple[Collection[Parameter], Mapping[str, AnyType]]] = None
-    required: InitVar[bool] = False
     resolve: Optional[Callable] = None
     schema: InitVar[Optional[Schema]] = None
     subscribe: Optional[Callable] = None
     deprecated: Optional[str] = field_(init=False, default=None)
     description: Optional[str] = field_(init=False, default=None)
 
-    def __post_init__(self, required: bool, schema: Optional[Schema]):
-        if required:
-            object.__setattr__(self, "default", graphql.Undefined)
+    def __post_init__(self, schema: Optional[Schema]):
         if schema is not None and schema.annotations is not None:
             object.__setattr__(self, "description", schema.annotations.description)
             if schema.annotations.deprecated is True:
@@ -143,7 +139,7 @@ class ObjectField:
                 )
             elif isinstance(schema.annotations.deprecated, str):
                 object.__setattr__(self, "deprecated", schema.annotations.deprecated)
-            if schema.annotations.default is not Undefined and not required:
+            if schema.annotations.default is not Undefined:
                 object.__setattr__(self, "default", schema.annotations.default)
 
 
@@ -208,31 +204,17 @@ class SchemaBuilder(ConversionsVisitor[Conv, Thunk[graphql.GraphQLType]]):
     def _object_field(self, field: Field, field_type: AnyType) -> ObjectField:
         field_name = field.name
         field_type, conversion = get_field_conversion(field, field_type, self.operation)
-        conversions: Optional[Conversions] = None
-        resolve: Optional[Callable] = None
+        converter, conversions = identity, None
         if conversion is not None:
-            converter, conversions = conversion.converter, conversion.conversions
-            aliaser = self.aliaser
-
-            def resolve(obj, info):
-                return partial_serialize(
-                    converter(getattr(obj, field_name)),
-                    conversions=conversions,
-                    aliaser=aliaser,
-                )
-
-        default: Any = graphql.Undefined
-        if not is_required(field):
-            with suppress(Exception):
-                default = serialize(get_default(field), conversions=conversions)
+            converter = conversion.converter  # type: ignore
+            conversions = conversion.conversions
         return ObjectField(
             field_name,
             field_type,
             alias=get_alias(field),
+            converter=converter,
             conversions=conversions,
-            default=default,
-            required=is_required(field),
-            resolve=resolve,
+            default=graphql.Undefined if is_required(field) else get_default(field),
             schema=field.metadata.get(SCHEMA_METADATA),
         )
 
@@ -294,20 +276,14 @@ class SchemaBuilder(ConversionsVisitor[Conv, Thunk[graphql.GraphQLType]]):
         types: Mapping[str, AnyType],
         defaults: Mapping[str, Any],
     ) -> Thunk[graphql.GraphQLType]:
-        fields = []
-        for field_name, field_type in types.items():
-            default = graphql.Undefined
-            if field_name in defaults:
-                with suppress(Exception):
-                    default = serialize(defaults[field_name])
-            fields.append(
-                ObjectField(
-                    field_name,
-                    field_type,
-                    default=default,
-                    required=field_name in defaults,
-                )
+        fields = [
+            ObjectField(
+                field_name,
+                field_type,
+                default=defaults.get(field_name, graphql.Undefined),
             )
+            for field_name, field_type in types.items()
+        ]
         return self.object(cls, fields)
 
     def new_type(self, tp: Type, super_type: AnyType) -> Thunk[graphql.GraphQLType]:
@@ -423,19 +399,28 @@ class InputSchemaBuilder(
     SchemaBuilder[Deserialization],
 ):
     def _field(self, field: ObjectField) -> Tuple[str, Lazy[graphql.GraphQLInputField]]:
-        type_thunk = self.visit_with_conversions(field.type, field.conversions)
+        field_type = field.type
+        default: Any = graphql.Undefined
+        # Don't put `null` default + handle Undefined as None
+        if field.default in {None, Undefined}:
+            field_type = Optional[field_type]
+        elif field.default is not graphql.Undefined:
+            try:
+                default = serialize(
+                    field.converter(field.default),
+                    conversions=field.conversions,
+                    aliaser=self.aliaser,
+                )
+            except Exception:
+                field_type = Optional[field_type]
+
+        type_thunk = self.visit_with_conversions(field_type, field.conversions)
 
         def field_thunk():
-            field_type = exec_thunk(type_thunk)
-            if (
-                not isinstance(field_type, graphql.GraphQLNonNull)
-                and field.default is None
-            ):
-                default = graphql.Undefined
-            else:
-                default = field.default
             return graphql.GraphQLInputField(
-                field_type, default_value=default, description=field.description
+                exec_thunk(type_thunk),
+                default_value=default,
+                description=field.description,
             )
 
         return self.aliaser(field.alias or field.name), field_thunk
@@ -502,16 +487,22 @@ class OutputSchemaBuilder(
         if field.resolve is not None:
             resolve = field.resolve
         else:
-            field_name, conv, aliaser = field.name, field.conversions, self.aliaser
+            field_name, aliaser = field.name, self.aliaser
+            converter, conversions = field.converter, field.conversions
 
             def resolve(obj, _):
                 return partial_serialize(
-                    getattr(obj, field_name), conversions=conv, aliaser=aliaser
+                    converter(getattr(obj, field_name)),
+                    aliaser=aliaser,
+                    conversions=conversions,
                 )
 
         resolve = self._wrap_resolve(resolve)
 
-        field_type = self.visit_with_conversions(field.type, field.conversions)
+        field_type = field.type
+        if is_union_of(field_type, UndefinedType):
+            field_type = Optional[field_type]
+        type_thunk = self.visit_with_conversions(field_type, field.conversions)
         args = None
         if field.parameters is not None:
             parameters, types = field.parameters
@@ -521,10 +512,13 @@ class OutputSchemaBuilder(
                 param_type = types[param.name]
                 if is_union_of(param_type, graphql.GraphQLResolveInfo):
                     break
-                # None because of https://github.com/python/typing/issues/775
-                if param.default in {None, Undefined, graphql.Undefined}:
+                # Don't put `null` default + handle Undefined as None
+                # also https://github.com/python/typing/issues/775
+                if param.default in {None, Undefined}:
                     param_type = Optional[param_type]
-                if param.default != Parameter.empty:
+                # param.default == graphql.Undefined means the parameter is required
+                # even if it has a default
+                elif param.default not in {Parameter.empty, graphql.Undefined}:
                     try:
                         default = serialize(param.default)
                     except Exception:
@@ -534,17 +528,11 @@ class OutputSchemaBuilder(
                     arg_thunk=self.input_builder.visit(param_type),
                     default=default,
                 ) -> graphql.GraphQLArgument:
-                    arg_type = exec_thunk(arg_thunk)
-                    if (
-                        not isinstance(arg_type, graphql.GraphQLNonNull)
-                        and default is None
-                    ):
-                        default = graphql.Undefined
-                    return graphql.GraphQLArgument(arg_type, default)
+                    return graphql.GraphQLArgument(exec_thunk(arg_thunk), default)
 
                 args[self.aliaser(param.name)] = arg_thunk
         return self.aliaser(field.alias or field.name), lambda: graphql.GraphQLField(
-            exec_thunk(field_type),
+            exec_thunk(type_thunk),
             {name: arg() for name, arg in args.items()} if args else None,
             resolve,
             field.subscribe,
@@ -567,8 +555,8 @@ class OutputSchemaBuilder(
         def get_merge(obj):
             return partial_serialize(
                 converter(getattr(get_prev_merged(obj), field_name)),
-                conversions=conversions,
                 aliaser=aliaser,
+                conversions=conversions,
             )
 
         merged_save = self._get_merged
@@ -690,14 +678,27 @@ class Operation(Generic[T]):
     error_handler: ErrorHandler = Undefined
 
 
-OpOrFunc = Union[Callable[..., T], Operation[T]]
+class Query(Operation):
+    pass
+
+
+class Mutation(Operation):
+    pass
+
+
+@dataclass(frozen=True)
+class Subscription(Operation[AsyncIterable]):
+    resolver: Optional[Callable] = None
+
+
+Op = TypeVar("Op", bound=Operation)
 
 
 def operation_resolver(
-    operation: OpOrFunc, *, keep_first_param=False
+    operation: Union[Callable, Op], op_class: Type[Op]
 ) -> Tuple[str, Resolver]:
-    if not isinstance(operation, Operation):
-        operation = Operation(operation)
+    if not isinstance(operation, op_class):
+        operation = op_class(operation)  # type: ignore
     error_handler: Optional[Callable]
     if operation.error_handler is Undefined:
         error_handler = None
@@ -705,53 +706,30 @@ def operation_resolver(
         error_handler = none_error_handler
     else:
         error_handler = operation.error_handler
-    if keep_first_param:
-        wrapper = operation.function
+    op = operation.function
+    if iscoroutinefunction(op):
+
+        async def wrapper(_, *args, **kwargs):
+            return await op(*args, **kwargs)
+
     else:
-        op = operation.function
-        if iscoroutinefunction(op):
 
-            async def wrapper(_, *args, **kwargs):
-                return await op(*args, **kwargs)
+        def wrapper(_, *args, **kwargs):
+            return op(*args, **kwargs)
 
-        else:
+    wrapper.__annotations__ = op.__annotations__
 
-            def wrapper(_, *args, **kwargs):
-                return op(*args, **kwargs)
-
-        wrapper.__annotations__ = op.__annotations__
-
-    (*parameters,) = resolver_parameters(
-        operation.function, check_first=not keep_first_param
-    )
-    if keep_first_param:
-        parameters = parameters[1:]
+    (*parameters,) = resolver_parameters(operation.function, check_first=True)
     return operation.alias or operation.function.__name__, Resolver(
         wrapper, operation.conversions, operation.schema, error_handler, parameters
     )
 
 
-def remove_error_handler(subscription: OpOrFunc) -> OpOrFunc:
-    if (
-        isinstance(subscription, Operation)
-        and subscription.error_handler is not Undefined
-    ):
-        warnings.warn("Subscriber error_handler is ignored")
-        return replace(subscription, error_handler=Undefined)
-    else:
-        return subscription
-
-
 def graphql_schema(
     *,
-    query: Iterable[OpOrFunc] = (),
-    mutation: Iterable[OpOrFunc] = (),
-    subscription: Iterable[
-        Union[
-            OpOrFunc[AsyncIterable],
-            Tuple[Callable[..., AsyncIterable], OpOrFunc],
-        ]
-    ] = (),
+    query: Iterable[Union[Callable, Query]] = (),
+    mutation: Iterable[Union[Callable, Mutation]] = (),
+    subscription: Iterable[Union[Callable[..., AsyncIterable], Subscription]] = (),
     types: Iterable[Type] = (),
     directives: Optional[Collection[graphql.GraphQLDirective]] = None,
     description: Optional[str] = None,
@@ -767,56 +745,77 @@ def graphql_schema(
     query_fields: List[ObjectField] = []
     mutation_fields: List[ObjectField] = []
     subscription_fields: List[ObjectField] = []
-    for operations, fields in [(query, query_fields), (mutation, mutation_fields)]:
-        for operation in operations:
-            name, resolver = operation_resolver(operation)
+    for operations, op_class, fields in [
+        (query, Query, query_fields),
+        (mutation, Mutation, mutation_fields),
+    ]:
+        for operation in operations:  # type: ignore
+            name, resolver = operation_resolver(operation, op_class)
             resolver_types = resolver.types()
             fields.append(
                 ObjectField(
                     name,
                     resolver_types["return"],
+                    conversions=resolver.conversions,
                     parameters=(resolver.parameters, resolver_types),
                     resolve=resolver_resolve(resolver, resolver_types, aliaser),
                     schema=resolver.schema,
                 )
             )
     for sub_op in subscription:  # type: ignore
-        resolve: Callable
-        if isinstance(sub_op, tuple):
-            operation, event_handler = cast(Tuple[Callable, OpOrFunc], sub_op)
-            operation = remove_error_handler(operation)
-            name, resolver = operation_resolver(event_handler, keep_first_param=True)
-            _, subscriber = operation_resolver(operation)
+        if not isinstance(sub_op, Subscription):
+            sub_op = Subscription(sub_op)  # type: ignore
+        sub_parameters: Sequence[Parameter]
+        if sub_op.resolver is not None:
+            name = sub_op.alias or sub_op.resolver.__name__
+            _, subscriber2 = operation_resolver(sub_op, Subscription)
+            _, *sub_parameters = resolver_parameters(sub_op.resolver, check_first=False)
+            resolver = Resolver(
+                sub_op.resolver,
+                sub_op.conversions,
+                sub_op.schema,
+                subscriber2.error_handler,
+                sub_parameters,
+            )
+            sub_types = resolver.types()
+            sub_return = sub_types["return"]
+            resolve = resolver_resolve(resolver, sub_types, aliaser)
+            subscriber = replace(subscriber2, error_handler=None)
             subscribe = resolver_resolve(
                 subscriber, subscriber.types(), aliaser, serialized=False
             )
-            resolver_types = resolver.types()
-            resolve = resolver_resolve(resolver, resolver_types, aliaser)
-            return_type = resolver_types["return"]
         else:
-            operation = remove_error_handler(cast(OpOrFunc, sub_op))
-            name, resolver = operation_resolver(sub_op)
-            resolver_types = resolver.types()
-            if get_origin2(resolver_types["return"]) not in async_iterable_origins:
+            name, subscriber2 = operation_resolver(sub_op, Subscription)
+            resolver = Resolver(
+                lambda _: _,
+                sub_op.conversions,
+                sub_op.schema,
+                subscriber2.error_handler,
+                (),
+            )
+            resolve = resolver_resolve(resolver, {}, aliaser)
+            subscriber = replace(subscriber2, error_handler=None)
+            sub_parameters = subscriber.parameters
+            sub_types = subscriber.types()
+            if get_origin2(sub_types["return"]) not in async_iterable_origins:
                 raise TypeError(
                     "Subscriptions must return an AsyncIterable/AsyncIterator"
                 )
-            return_type = get_args2(resolver_types["return"])[0]
+            event_type = get_args2(sub_types["return"])[0]
             subscribe = resolver_resolve(
-                resolver, resolver_types, aliaser, serialized=False
+                subscriber, sub_types, aliaser, serialized=False
             )
-
-            def resolve(_, *args, **kwargs):
-                return _
+            sub_return = resolver.return_type(event_type)
 
         subscription_fields.append(
             ObjectField(
                 name,
-                return_type,
-                parameters=(resolver.parameters, resolver_types),
+                sub_return,
+                conversions=sub_op.conversions,
+                parameters=(sub_parameters, sub_types),
                 resolve=resolve,
                 subscribe=subscribe,
-                schema=resolver.schema,
+                schema=sub_op.schema,
             )
         )
 
