@@ -1,3 +1,4 @@
+from contextlib import suppress
 from dataclasses import Field
 from enum import Enum
 from typing import (  # type: ignore
@@ -7,6 +8,7 @@ from typing import (  # type: ignore
     Mapping,
     Optional,
     Sequence,
+    Set,
     Tuple,
     Type,
 )
@@ -16,7 +18,6 @@ from apischema.dataclass_utils import get_field_conversion, get_fields
 from apischema.json_schema.refs import get_ref, schema_ref
 from apischema.skip import filter_skipped
 from apischema.types import AnyType
-from apischema.utils import is_hashable
 
 try:
     from apischema.typing import Annotated
@@ -34,30 +35,31 @@ class RefsExtractor(ConversionsVisitor):
     def __init__(self, refs: Refs):
         super().__init__()
         self.refs = refs
+        self._rec_guard: Set[AnyType] = set()
 
-    def _incr_ref(self, ref: Optional[str], cls: AnyType) -> bool:
+    def _incr_ref(self, ref: Optional[str], tp: AnyType) -> bool:
         if ref is None:
             return False
         else:
-            ref_cls, count = self.refs.get(ref, (cls, 0))
-            if ref_cls != cls:
+            ref_cls, count = self.refs.get(ref, (tp, 0))
+            if ref_cls != tp:
                 raise ValueError(
-                    f"Types {cls} and {self.refs[ref]} share same reference '{ref}'"
+                    f"Types {tp} and {self.refs[ref]} share same reference '{ref}'"
                 )
             self.refs[ref] = (ref_cls, count + 1)
             return count > 0
 
-    def annotated(self, cls: AnyType, annotations: Sequence[Any]):
+    def annotated(self, tp: AnyType, annotations: Sequence[Any]):
         for annotation in reversed(annotations):
             if isinstance(annotation, schema_ref):
-                annotation.check_type(cls)
+                annotation.check_type(tp)
                 ref = annotation.ref
                 if not isinstance(ref, str):
                     raise ValueError("Annotated schema_ref can only be str")
-                annotated = Annotated[(cls, *filter(is_hashable, annotations))]  # type: ignore # noqa: E501
+                annotated = Annotated[(tp, *annotations)]  # type: ignore
                 if self._incr_ref(ref, annotated):
                     return
-        return self.visit(cls)
+        return self.visit(tp)
 
     def any(self):
         pass
@@ -73,11 +75,13 @@ class RefsExtractor(ConversionsVisitor):
         init_vars: Sequence[Field],
     ):
         for field in get_fields(fields, init_vars, self.operation):
-            field_type, conversions, _ = get_field_conversion(
+            field_type, conversion = get_field_conversion(
                 field, types[field.name], self.operation
             )
-            with self._replace_conversions(conversions):
-                self.visit(field_type)
+            self.visit_with_conversions(
+                field_type,
+                conversion.conversions if conversion is not None else None,
+            )
 
     def enum(self, cls: Type[Enum]):
         pass
@@ -98,7 +102,7 @@ class RefsExtractor(ConversionsVisitor):
         for cls in types.values():
             self.visit(cls)
 
-    def new_type(self, cls: AnyType, super_type: AnyType):
+    def new_type(self, tp: AnyType, super_type: AnyType):
         self.visit(super_type)
 
     def primitive(self, cls: Type):
@@ -120,14 +124,21 @@ class RefsExtractor(ConversionsVisitor):
             pass
 
     def union(self, alternatives: Sequence[AnyType]):
-        return self._union_result(
-            map(self.visit, filter_skipped(alternatives, schema_only=True))
-        )
+        for tp in filter_skipped(alternatives, schema_only=True):
+            self.visit(tp)
 
-    def visit(self, cls: AnyType):
-        if (
-            not is_hashable(cls)
-            or self.is_extra_conversions(cls)
-            or not self._incr_ref(get_ref(cls), cls)
-        ):
-            super().visit(cls)
+    def visit(self, tp: AnyType):
+        dynamic = self._apply_dynamic_conversions(tp)
+        ref_tp = dynamic if dynamic is not None else tp
+        with suppress(TypeError):
+            recursion = ref_tp in self._rec_guard
+        if not self._incr_ref(get_ref(ref_tp), ref_tp):
+            if recursion:
+                raise TypeError(f"Recursive type {tp} need a ref")
+            with suppress(TypeError):
+                self._rec_guard.add(ref_tp)
+            try:
+                super().visit(ref_tp)
+            finally:
+                with suppress(TypeError):
+                    self._rec_guard.discard(ref_tp)
