@@ -13,7 +13,7 @@ from typing import (
 )
 
 from apischema.types import AnyType, MetadataMixin, Number
-from apischema.utils import Undefined, merge_opts
+from apischema.utils import Undefined, merge_opts, replace_builtins
 from .annotations import Annotations, Deprecated, merge_annotations
 from .constraints import (
     ArrayConstraints,
@@ -23,27 +23,32 @@ from .constraints import (
     StringConstraints,
     merge_constraints,
 )
-from ..typing import get_origin
+from apischema.metadata.keys import SCHEMA_METADATA
+from apischema.typing import get_origin
+
+try:
+    from apischema.typing import Annotated
+except ImportError:
+    Annotated = ...  # type: ignore
 
 T = TypeVar("T")
 
 
 @dataclass(frozen=True)
 class Schema(MetadataMixin):
+    key = SCHEMA_METADATA
     annotations: Optional[Annotations] = None
     constraints: Optional[Constraints] = None
     override: bool = False
 
-    def __post_init__(self):
-        from apischema.metadata.keys import SCHEMA_METADATA
+    def __call__(self, tp: T) -> T:
+        if get_origin(tp) is Annotated:
+            raise TypeError("Cannot register schema on Annotated type")
+        _schema[replace_builtins(tp)] = self
+        return tp
 
-        super().__init__(SCHEMA_METADATA)
-
-    def __call__(self, obj: T) -> T:
-        if get_origin(obj) is not None:
-            raise TypeError("Generic alias cannot have schema")
-        _schema[obj] = self
-        return obj
+    def __set_name__(self, owner, name):
+        self.__call__(owner)
 
     def as_dict(self) -> Mapping[str, Any]:
         result: Dict[str, Any] = {}
@@ -165,15 +170,16 @@ def schema(override=False, **kwargs) -> Schema:
     return Schema(annotations, constraints, override)
 
 
-def _default_schema(cls: AnyType) -> Optional[Schema]:
+def _default_schema(tp: AnyType) -> Optional[Schema]:
     return None
 
 
 _schema: Dict[Any, Schema] = {}
 
 
-def get_schema(cls: AnyType) -> Optional[Schema]:
-    return _schema[cls] if cls in _schema else _default_schema(cls)
+def get_schema(tp: AnyType) -> Optional[Schema]:
+    tp = replace_builtins(tp)
+    return _schema[tp] if tp in _schema else _default_schema(tp)
 
 
 @merge_opts
