@@ -1,6 +1,7 @@
 import collections.abc
 import sys
 import typing
+from itertools import chain
 from types import MappingProxyType
 from typing import (
     AbstractSet,
@@ -60,15 +61,15 @@ MAPPING_TYPES = {
 
 if (3, 7) <= sys.version_info < (3, 9):  # pragma: no cover
 
-    def subscriptable_origin(cls: AnyType) -> AnyType:
+    def subscriptable_origin(tp: AnyType) -> AnyType:
         if (
-            type(cls) == type(List[int])  # noqa: E721
-            and cls.__module__ == "typing"
-            and hasattr(cls, "_name")
+            type(tp) == type(List[int])  # noqa: E721
+            and tp.__module__ == "typing"
+            and hasattr(tp, "_name")
         ):
-            return getattr(typing, cls._name)
+            return getattr(typing, tp._name)
         else:
-            return get_origin(cls)
+            return get_origin(tp)
 
 
 else:  # pragma: no cover
@@ -77,8 +78,13 @@ else:  # pragma: no cover
 
 if sys.version_info >= (3, 7):  # pragma: no cover
     OrderedDict = dict
+    ChainMap = collections.ChainMap
 else:  # pragma: no cover
-    from collections import OrderedDict  # noqa
+    OrderedDict = collections.OrderedDict
+
+    class ChainMap(collections.ChainMap):
+        def __iter__(self):
+            return iter({k: None for k in chain.from_iterable(reversed(self.maps))})
 
 
 class MetadataUnion(Mapping[str, Any]):
@@ -97,18 +103,15 @@ else:  # pragma: no cover
 
 
 class MetadataMixin(MetadataUnion):
-    _key: str
-
-    def __init__(self, key: str):
-        super().__setattr__("_key", key)
+    key: str
 
     def __getitem__(self, key):
-        if key != self._key:
+        if key != self.key:
             raise KeyError(key)
         return self
 
     def __iter__(self):
-        return iter((self._key,))
+        return iter((self.key,))
 
     def __len__(self):
         return 1
