@@ -2,7 +2,6 @@ __all__ = [
     "Discard",
     "ValidationError",
     "ValidatorResult",
-    "add_validator",
     "get_validators",
     "validate",
     "validator",
@@ -10,4 +9,4 @@ __all__ = [
 ]
 
 from .errors import ValidationError, ValidatorResult, with_validation_error
-from .validator import Discard, add_validator, get_validators, validate, validator
+from .validator import Discard, get_validators, validate, validator
