@@ -1,11 +1,10 @@
-from dataclasses import MISSING, dataclass
+from dataclasses import dataclass, fields
 from functools import partial
 from types import FunctionType, MethodType
 from typing import Any, Mapping, Optional, TYPE_CHECKING, Type, TypeVar
 
-from apischema.dataclasses import fields_items
-from apischema.fields import FIELDS_SET_ATTR
-from apischema.utils import get_default
+from apischema.dataclass_utils import get_default, has_default
+from apischema.fields import FIELDS_SET_ATTR, set_fields
 
 if TYPE_CHECKING:
     from apischema.validation.validator import Validator
@@ -20,39 +19,38 @@ class NonTrivialDependency(Exception):
         self.validator: Optional["Validator"] = None
 
 
-@dataclass
+@dataclass(init=False)
 class ValidatorMock:
-    def __init__(self, cls: Type, fields: Mapping[str, Any]):
+    def __init__(self, cls: Type, values: Mapping[str, Any]):
         self.cls = cls
-        self.fields = fields
-        setattr(self, FIELDS_SET_ATTR, set(fields))
+        self.values = values
+        set_fields(self, *values, overwrite=True)
 
     def __getattribute__(self, name: str) -> Any:
-        fields = super().__getattribute__("fields")
-        if name in fields:
-            return fields[name]
+        values = super().__getattribute__("values")
+        if name in values:
+            return values[name]
         cls = super().__getattribute__("cls")
-        cls_fields = fields_items(cls)
-        if name in cls_fields:
-            try:
-                return get_default(cls_fields[name])
-            except NotImplementedError:
-                raise NonTrivialDependency(name) from None
+        for field in fields(cls):
+            if name == field.name:
+                try:
+                    return get_default(field)
+                except NotImplementedError:
+                    raise NonTrivialDependency(name) from None
         if name == "__class__":
             return cls
         if name == "__dict__":
             return {
-                **fields,
+                **values,
                 **{
-                    name: get_default(field)
-                    for name, field in cls_fields.items()
-                    if field.default is not MISSING
-                    or field.default_factory is not MISSING  # type: ignore
+                    field.name: get_default(field)
+                    for field in fields(cls)
+                    if has_default(field)
                 },
-                FIELDS_SET_ATTR: set(fields),
+                FIELDS_SET_ATTR: set(values),
             }
         if name == FIELDS_SET_ATTR:
-            return set(fields)
+            return set(values)
         if hasattr(cls, name):
             member = getattr(cls, name)
             # for classmethod (staticmethod are not handled)
